//Arduino JWT lib used: https://github.com/yutter/ArduinoJWT
//Target device for programming is NodeMCU-32S, but having the hardware is not necessary to solve this challenge.

#include <WiFi.h>
#include <WiFiclient.h>
#include <WiFiAP.h>
#include "ArduinoJWT.h"
#include "Defines.h"

boolean DEBUG_ENABLED = true;

WiFiServer server(SERVER_PORT); 
ArduinoJWT jwtHandler(JWT_SYMMETRIC_KEY_FOR_SIGNATURE); 


void convertCharArrayContainingHexaValuesAsASCIIToByteArray(const char *input, unsigned int inputLength, char *result) {
  for (int i = 0; i < inputLength / 2; i++) { 
    char a = input[i * 2]; 
    char b = input[i * 2 + 1];

    a = (a <= '9') ? a - '0' : (a & 0x7) + 9; 
    b = (b <= '9') ? b - '0' : (b & 0x7) + 9;

    result[i] = (a << 4) + b; 
  }
}


void makeResponse(WiFiClient clientRequest, int httpStatus, String message = "") {
  switch (httpStatus) {
    case 400:
      if (DEBUG_ENABLED) Serial.print("Response: HTTP/1.1 400 Bad Request");
      clientRequest.println("HTTP/1.1 400 Bad Request");
      clientRequest.println("Connection: close");
      clientRequest.println();
      break;

    case 401:
      if (DEBUG_ENABLED) Serial.print("Response: HTTP/1.1 401 Unauthorized");
      clientRequest.println("HTTP/1.1 401 Unauthorized");
      clientRequest.println("WWW-Authenticate: Bearer");
      clientRequest.println("Connection: close");
      clientRequest.println();
      break;

    case 403:
      if (DEBUG_ENABLED) Serial.print("Response: HTTP/1.1 403 Forbidden");
      clientRequest.println("HTTP/1.1 403 Forbidden");
      clientRequest.println("Connection: close");
      clientRequest.println();
      break;

    case 200:
      if (DEBUG_ENABLED) {
        Serial.print("Response: HTTP/1.1 200 OK -> ");
        Serial.println(message);
      }
      clientRequest.println("HTTP/1.1 200 OK");
      clientRequest.println("Content-type:text/html");
      clientRequest.println("Connection: close");
      clientRequest.println();
      clientRequest.println(message);
      clientRequest.println();
  }
}


String readRequest(WiFiClient clientRequest) {
  boolean lastCharacterWasANewLine = false; 
  String request = "";

  while (clientRequest.connected()) { 
    if (clientRequest.available()) { 
      char c = clientRequest.read(); 
      request += c; 

      if (c == '\n') { 
        if (lastCharacterWasANewLine) {
          break;
        } else {
          lastCharacterWasANewLine = true;
        }
      } else if (c != '\r') { 
        lastCharacterWasANewLine = false;
      }
    }
  }

  request.trim(); 
  return request;
}


void setup() {
  pinMode(LED_BUILTIN, OUTPUT); 

  if (DEBUG_ENABLED) Serial.begin(SERIAL_BAUDRATE); 
  if (DEBUG_ENABLED) Serial.println();

  if (DEBUG_ENABLED) Serial.println("Configuring access point...");
  WiFi.mode(WIFI_AP); 
  WiFi.softAP(WIFI_SSID); 
  IPAddress myIP = WiFi.softAPIP(); 

  if (DEBUG_ENABLED) {
    Serial.print("AP IP address: ");
    Serial.println(myIP);
  }

  server.begin(); 

  if (DEBUG_ENABLED) Serial.println("Server started");
}


void loop() {
  WiFiClient clientRequest = server.available();   

  if (clientRequest) { 
    digitalWrite(LED_BUILTIN, HIGH); 
    if (DEBUG_ENABLED) Serial.println("\n\nNew request:");
    String request = readRequest(clientRequest); 

    if (DEBUG_ENABLED) {
      Serial.println(request);
      Serial.println();
    }

    int indexOfStartOfBearerToken = request.indexOf("Bearer "); 
    if (indexOfStartOfBearerToken >= 0) { 
      String jwtToken = request.substring(indexOfStartOfBearerToken + 7, request.length()); 
      jwtToken = jwtToken.substring(0, jwtToken.indexOf("\n")); 
      jwtToken.trim(); 

      String jwtPayload = "";
      if (jwtHandler.decodeJWT(jwtToken, jwtPayload)) { 
        if (jwtPayload.equals(MASTER_JWT_PAYLOAD)) { 
          String seed = jwtPayload.substring(jwtPayload.indexOf("\"magic_values\":\"") + 16, jwtPayload.length()); 
          seed = seed.substring(0, seed.indexOf("\""));

          int seedLength = seed.length();
          char seedConvertedToInt[seedLength / 2]; 
          convertCharArrayContainingHexaValuesAsASCIIToByteArray(seed.c_str(), seedLength, seedConvertedToInt); 

          char flagBasevaluesConvertedToInt[seedLength / 2];
          convertCharArrayContainingHexaValuesAsASCIIToByteArray(FLAG_BASEVALUES, seedLength, flagBasevaluesConvertedToInt); 

          String responseString = "<html><body><h2>Welcome back, Admin!</h2><p>Your entry code for today is: "; 

          for (int i = 0; i < seedLength / 2; i++) {
            responseString += (char)(seedConvertedToInt[i] ^ flagBasevaluesConvertedToInt[i]);  
          }

          responseString += "</p><p>Hope you have a good day, please be sure to leave a feedback to our new super cool Auth system!</p></body></html>"; 

          makeResponse(clientRequest, 200, responseString); 
        } else {
          makeResponse(clientRequest, 403); 
        }
      } else {
        makeResponse(clientRequest, 400); 
      }
    } else {
      makeResponse(clientRequest, 401); 
    }

    clientRequest.stop(); 
    if (DEBUG_ENABLED) Serial.println("\nclientRequest disconnected.\n\n\n");
    digitalWrite(LED_BUILTIN, LOW); 
  }
}
